:
# NAME:
#	dns_convzone.sh - DNS zone file and config conversion
#
# SYNOPSIS:
#	dns_convzone.sh < named.boot
#
# DESCRIPTION:
#	Routines used by dns_convert.sh, also provides a standalone
#	mechanism to convert zone files (just provide input in
#	named.boot format).
#
# SEE ALSO:
#	dns_convert.sh(8), ndconv.sh(8)
#	
# AUTHOR:
#	Simon J. Gerraty <sjg@crufty.net>
#

# RCSid:
#	$Id: dns_convzone.sh,v 1.11 2022/10/10 00:39:06 sjg Exp $
#
#	@(#) Copyright (c) 1997-2022 Simon J. Gerraty.
#
#	This file is provided in the hope that it will
#	be of use.  There is absolutely NO WARRANTY.
#	Permission to copy, redistribute or otherwise
#	use this file is hereby granted provided that 
#	the above copyright notice and this notice are
#	left intact. 
#      
#	Please send copies of changes and bug-fixes to:
#	sjg@crufty.net
#

Mydir=${Mydir:-`dirname $0`}

PZONES=${PZONES:-primary.zones}
SZONES=${SZONES:-secondary.zones}

case "$PZONES" in
*zones) BINDVERS=${BINDVERS:-8};;
*)	BINDVERS=${BINDVERS:-4};;
esac

case "$BINDVERS" in
4*)	BIND4=
	BIND8=:
	;;
*)	BIND4=:
	BIND8=
	;;
esac

case "$PRIMARY$SECONDARY" in
"")
	SECONDARY=:
	PRIMARY=
	;;
esac

mkdir -p hosts rev secondary mx

strip_soa() {
	cat <<'!'
;; RCSid:
;;	$Id: dns_convzone.sh,v 1.11 2022/10/10 00:39:06 sjg Exp $
;;
!
	sed -e '/IN.*SOA/,/)/d' -e '/^$INCLUDE[ 	]*ns.list[ 	]*$/d'
}

rev_zonefile() {
	datafile=${1:-rev}/
	set -- `IFS="."; echo $domain`
	i=`expr $# - 2`
	while [ $i -gt 0 ]
	do
		eval x=\$$i
		case "$x" in
		[0-9]*)	datafile="$datafile$x.";;
		esac
		i=`expr $i - 1`
	done
	soafile=${datafile}soa
	datafile=${datafile}db
}

mkzone() {
	type=$1
	domain=$2
	file=$3
	shift 3
	primaries=
	case "$type" in
	primary|hint)
		;;
	*)
		for a in $*
		do
			primaries="$primaries $a;"
		done
		;;
	esac
	cat <<!EOZ!
	
zone "$domain" {
	type $type;
	file "$file";
!EOZ!
	test "$primaries" && echo "	primaries { $primaries };"
	echo "};"
	echo
}

mkhead() {
	echo "// generated by $0" 
	cat <<'!EOH!'
//
// RCSid:
//	$Id: dns_convzone.sh,v 1.11 2022/10/10 00:39:06 sjg Exp $
//
!EOH!
}

mkrc() {
	grep -i 'IN.*SOA' | sed -e 's,.*SOA\(.*\)(.*,\1,' -e 's/_/-/g' -e 's/\.[	]/ /g' |
	while read ns email
	do
		cat <<! > .dnsrc.pl
# generated by $0		 
\$opt_n = '$ns';
\$opt_M = '$email';
1;
!
		cat <<! > .dnsrc
# generated by $0		 
opt_n='$ns'
opt_M='$email'
!
	done
}

conv_zones() {
	# input is named.boot format
	while read key domain rest
	do
		if test "$BIND4"; then	# Not BIND-4
			case "$key" in
			round_robin*)
				n=`IFS=" ="; set -- $key$domain$rest; echo $2`
				echo "	round-robin ${n:-1};" >> $OPTS
				continue
				;;
			forwarders)
				echo "	forwarders {" >> $OPTS
				for a in $domain $rest
				do
					echo "		$a;"
				done >> $OPTS
				echo "	};" >> $OPTS
				continue
				;;
			options)
				echo "// $key $domain $rest" >> $OPTS
				for o in $domain $rest
				do
					case $o in
					forward-only) echo "	forward only;" >> $OPTS;;
					esac
				done
				continue
				;;
			secondary)
				echo "	forward only;" >> $OPTS
				continue
				;;
			"")
				echo
				continue
				;;
			\;*)
				echo "//$key $domain $rest"
				continue
				;;
			esac
		fi
		case "$key" in
		directory)
			dir=$DESTDIR$domain
			# since we are in a re-directed loop
			# variable settings will be lost
			# store any we need in .env
			echo "dir=$dir" >> .env
			$BIND4 echo "$key ${DNSDIR:-/var/named}"
			continue
			;;
		primary)
			test -s .dnsrc.pl || mkrc < $dir/$rest
			domain=`echo $domain | tr 'A-Z' 'a-z'`

			case "$rest" in
			db.auth)
				# This is magic, it means we've not yet
				# configured anything, we are just claiming
				# the domain.  We leave it alone as it is
				# instructive in named.boot.
				$BIND4 echo "$key $domain $rest"
				$BIND8 mkzone primary $domain $rest >> $PZONES
				continue
				;;
			*.soa)
				# already converted
				b=`echo $rest | sed 's,\.soa$,,'`
				case "$domain" in
				*arpa) datafile=$b.rev;;
				*) datafile=$b.db;;
				esac
				cp $dir/$datafile $datafile
				$BIND8 mkzone primary $domain $rest >> $PZONES
				$BIND4 echo "$key $domain $soafile"
				continue
				;;
			esac
			case "$domain" in
			.)	soafile=root.soa
                                datafile=root.db
                                ;;
#			127.in-addr.arpa|0.0.127.in-addr.arpa)
#				soafile=named.local
#				datafile=/dev/null
#				;;
			*arpa)
				rev_zonefile rev
				;;
			*)
				soafile="hosts/$domain.soa"
				datafile="hosts/$domain.db"
                                ;;
			esac
			$BIND4 echo "$key $domain $soafile"
			case "$domain" in
			.)	$BIND8 mkzone primary $domain $soafile
                                ;;
			127.in-addr.arpa|0.0.127.in-addr.arpa)
				$BIND8 echo "// every server primary for this"
				$BIND8 mkzone primary $domain $soafile
				;;
			*)
				$BIND8 mkzone primary $domain $soafile >> $PZONES
				;;
			esac
			case "$PRIMARY" in
			"")
				strip_soa < $dir/$rest | grep -iv "origin[ 	][ 	]*$domain" > $datafile
				;;
			esac
			continue
			;;
		secondary)
			domain=`echo $domain | tr 'A-Z' 'a-z'`

			case "$domain" in
			notyet*arpa)
				rev_zonefile secondary
				;;
			*)
				datafile="secondary/$domain";;
			esac
			set -- $rest
			server=
			while [ $# -gt 1 ]
			do
				server="$server$1 "
				shift
			done
			rest=$1
			$BIND4 echo "$key $domain $server $datafile"
			$BIND8 mkzone secondary $domain $datafile $server >> $SZONES
			$SECONDARY cp $dir/$rest $datafile
			continue
			;;
		cache)
			$PRIMARY cp $dir/$rest .
			$BIND8 mkzone hint . $rest
			$BIND8 continue
			;;
		esac
		$BIND4 echo "$key $domain $rest"
		$BIND8 echo "// $key $domain $rest"
	done
}

case "$0" in
*convzone*)
	conv_zones
	;;
esac
